@@ -10,6 +10,7 @@ package BasicLyX;
 use strict;
 
 use RelyxTable; # Handle LaTeX tables
+use RelyxFigure; # Handle LaTeX figures
 use Verbatim;   # Copy stuff verbatim
 
 use vars qw($bibstyle_insert_string $bibstyle_file $Begin_Inset_Include);
@@ -30,6 +31,13 @@ my $debug_on; # is debugging on?
 my $IsNewParagraph = 1;
 my $OldINP; #Save $IsNewParagraph during footnotes
 
+# Some layouts may have no actual text in them before the next layout
+# (e.g. slides). Pending Layout is set when we read a command that puts us
+# in a new layout. If we get some regular text to print out, set it to false.
+# But if we get to another layout command, first print out the command to
+# start the pending layout.
+my $PendingLayout = 0;
+
 # HACK to protect spaces within optional argument to \item
 my $protect_spaces = 0;
 
@@ -150,12 +158,14 @@ my %TextTokenTransTable = (
     '\LaTeXe' => "LaTeX2e",
     '\TeX'    => "TeX",
     '\LyX'    => "LyX",
+    '\lyxarrow' => "\\SpecialChar \\menuseparator\n",
     '\hfill'  => "\n\\hfill \n",
     '\noindent'        => "\n\\noindent \n",
     '\textbackslash'   => "\n\\backslash \n",
     '\textgreater'     => ">",
     '\textless'        => "<",
     '\textbar'         => "|",
+    '\textasciitilde'  => "~",
 );
 
 # LyX translations of some plain LaTeX text (TeX parser won't recognize it
@@ -178,10 +188,14 @@ my @LatexCommands = map {"\\$_"} qw(ref pageref label cite bibliography
 	                         index printindex tableofcontents
                                  listofalgorithms listoftables listoffigures);
 my @IncludeCommands = map {"\\$_"} qw(input include);
+# Included postscript files
+# LyX 1.0 can't do \includegraphics*!
+my @GraphicsCommands = map {"\\$_"} qw(epsf epsffile epsfbox
+                                       psfig epsfig includegraphics);
 
 # Accents. Most of these take an argument -- the thing to accent
 # (\l and \L are handled as InsetLatexAccents, so they go here too)
-my $AccentTokens = "\\\\[`'^#~=.uvHtcdblL\"]";
+my $AccentTokens = "\\\\[`'^#~=.bcdHklLrtuv\"]";
 
 # Environments which describe justifying (converted to LyX \align commands)
 #    and the corresponding LyX commands
@@ -200,7 +214,7 @@ my %ExtraArgEnvironments = (
 );
 
 # Math environments are copied verbatim
-my $MathEnvironments = "(math|equation|displaymath|eqnarray(\\*)?)";
+my $MathEnvironments = "(math|displaymath|xxalignat|(equation|eqnarray|align|alignat|xalignat|multline|gather)(\\*)?)";
 # ListLayouts may have standard paragraphs nested inside them.
 my $ListLayouts = "Itemize|Enumerate|Description";
 
@@ -474,6 +488,31 @@ sub basic_lyx {
 		    $thistable->hcline;
 		} else {warn "\\hline is illegal outside a table!"}
 
+	    # Figures
+
+	    } elsif ($name =~ /^\\epsf[xy]size$/) {
+		# We need to eat '=' followed by EITHER more text OR
+		# one (or more?!) macros describing a TeX size
+	        my $arg = $fileobject->eatMultiToken;
+		my $length = $arg->print;
+		$length =~ s/^\s*=\s*// or warn "weird '$name' command!";
+
+		# If there's no "cm" or other letters in $length, the next token
+		# ought to be something like \textwidth. Then it will be empty
+		# or just have numbers in it.
+		# This is bugprone. Hopefully not too many people use epsf!
+		if ($length =~ /^[\d.]*\s*$/) {
+		    my $next = $fileobject->eatMultiToken;
+		    $length .= $next->print;
+		}
+		$length =~ s/\s*$//; # may have \n at end
+
+		# If we can't parse the command, print it in tex mode
+		&RelyxFigure::parse_epsfsize($name, $length) or 
+		    &print_tex_mode("$name=$length");
+
+	    # Miscellaneous...
+
 	    } elsif ($name =~ /\\verb.*?/) {
 	        my $dummy = &Verbatim::copy_verb($fileobject, $eaten);
 		print "\nCopying $name in TeX mode: " if $debug_on;
@@ -539,7 +578,7 @@ sub basic_lyx {
 	    # This is to handle cases where _ is used, say, in a filename.
 	    # When _ is used in math mode, it'll be copied by the math mode
 	    # copying subs. Here we handle cases where it's used in non-math.
-	    # Examples are filenames for \includegraphics & citation labels.
+	    # Examples are filenames for \include & citation labels.
 	    # (It's illegal to use it in regular LaTeX text.)
 	    if ($name eq "_") {
 	       print OUTFILE $eaten->exact_print;
@@ -614,7 +653,29 @@ sub basic_lyx {
 
 		print OUTFILE "$pre_space\n",'\i ',$name,'{'
 	    
-	    # Miscellaneous commands
+	    # Included Postscript Figures
+	    # Currently, all postscript including commands take one
+	    # required argument and 0 to 2 optional args, so we can
+	    # clump them together in one else.
+	    } elsif (grep {$_ eq $name} @GraphicsCommands) {
+		&CheckForNewParagraph; # may be at the beginning of a par
+		my $arg1 = $fileobject->eatOptionalArgument;
+		# arg2 is a token of an empty string for most commands
+		my $arg2 = $fileobject->eatOptionalArgument;
+		my $arg3 = $fileobject->eatRequiredArgument;
+		my $save = $arg1->exact_print . $arg2->exact_print .
+		           $arg3->exact_print;
+
+		# Parse and put figure into LyX file
+		# Print it verbatim if we didn't parse correctly
+		my $thisfig = new RelyxFigure::Figure;
+		if ($thisfig->parse_pscommand($name, $arg1, $arg2, $arg3)) {
+		    print OUTFILE $thisfig->print_info;
+		} else {
+		    &print_tex_mode($eaten->exact_print . $save);
+		}
+
+	    # Tables
 
 	    } elsif ($name eq "\\multicolumn") {
 		if ($thistable = &RelyxTable::in_table) {
@@ -632,6 +693,8 @@ sub basic_lyx {
 		    $thistable->hcline($dummy);
 		} else {warn "\\cline is illegal outside a table!"}
 
+	    # Bibliography
+
 	    } elsif ($name eq '\bibliographystyle') {
 		$tok = $fileobject->eatRequiredArgument;
 		$bibstyle_file = "";
@@ -654,6 +717,8 @@ sub basic_lyx {
 		$tok = $fileobject->eatOptionalArgument;
 	        print OUTFILE "$name ", $tok->exact_print, "{";
 
+	    # Miscellaneous
+
 	    # ensuremath -- copy verbatim until you're done
 	    # but add \( and \)
 	    # Note that we'll only get here if the command is NOT in math mode
@@ -808,15 +873,16 @@ sub basic_lyx {
 		# start new pars). Then either print the extra arg or
 		# ignore it (depending on the environment).
 		if (exists $ExtraArgEnvironments{$env}) {
+		    # Should be just one token in the arg.
+		    my $arg = $fileobject->eatBalanced->contents->print;
+
 		    if ($ExtraArgEnvironments{$env}) { #print it
-			# Should be just one token in the arg.
-			my $arg = $fileobject->eatBalanced->contents->print;
 			print "\nArgument $arg to $env environment"
 								if $debug_on;
 			$item_preface = $ExtraArgEnvironments{$env} . $arg."\n";
 
 		    } else { #ignore it
-			print "\nIgnoring argument to $env environment"
+			print "\nIgnoring argument '$arg' to $env environment"
 								if $debug_on;
 		    }
 		} # end if for reading extra args to \begin command
@@ -845,13 +911,17 @@ sub basic_lyx {
 
 	    # Environments lyx translates to floats
 	    } elsif (exists $FloatEnvTransTable{$env}) {
+		# this really only matters if it's at the very
+		# beginning of the doc.
+		&CheckForNewParagraph;
+
 		$tok = $fileobject->eatOptionalArgument;
 		if ($tok && defined ($dummy = $tok->print) && $dummy) {
 		    print "\nIgnoring float placement '$dummy'" if $debug_on;
 		}
 	        my $command = $FloatEnvTransTable{$env};
 
-		# Open the footnote
+		# Open the table/figure
 		print OUTFILE "$command";
 
 	    # table
@@ -1222,8 +1292,8 @@ sub copy_latex_known {
 #    The goal of this subroutine is to create a block of LyX which will be
 # translated exactly back to the original when LyX creates its temporary LaTeX
 # file prior to creating a dvi file.
-#    Don't print \n\n at the end of the string... instead just set the
-# new paragraph flag. Also, don't print \n at the beginning of the string.
+#    Don't print \n\n at the end of the string... instead just set the new
+# paragraph flag. Also, don't print whitespace at the beginning of the string.
 # Print nothing if it's the beginning of a paragraph, or space otherwise.
 # These two things avoid extra C-Enter's in the LyX file
 sub print_tex_mode {
@@ -1231,12 +1301,13 @@ sub print_tex_mode {
 
     print "'$outstr'" if $debug_on;
 
-    # Handle extra \n's
+    # Handle extra \n's (& spaces) at beginning & end of string
     my $str_ends_par = ($outstr =~ s/\n{2,}$//);
     if ($IsNewParagraph) {
-        $outstr =~ s/\n//;  # .e.g, $outstr follows '\begin{quote}\n'
+        $outstr =~ s/^\s+//;  # .e.g, $outstr follows '\begin{quote}'
     } else {
-        $outstr =~ s/\n/ /; # e.g. $outstr follows '\LaTeX{}\n'
+	# Any whitespace is equivalent to one space in LaTeX
+        $outstr =~ s/^\s+/ /; # e.g. $outstr follows "\LaTeX{}" or "Hi{}"
     }
 
     # Check whether string came right at the beginning of a new paragraph
@@ -1298,6 +1369,9 @@ sub CheckForNewParagraph {
 	# Now that we've written the command, it's no longer a new paragraph
 	$IsNewParagraph = 0;
 
+	# And we're no longer waiting to see if this paragraph is empty
+	$PendingLayout = 0;
+
 	# When you write a new paragraph, reprint any font commands
 	foreach $dummy (keys %$CurrentFontStatus) {
 	    my $currf = $CurrentFontStatus->{$dummy}->[-1];
@@ -1321,6 +1395,7 @@ sub ConvertToLayout {
 
     my $layoutref = $ReadCommands::ToLayout->{$name};
     my $layout = $layoutref->{'layout'};
+    my $keepempty = $layoutref->{'keepempty'};
     my $dummy = ($name =~ /^\\/ ? "macro" : "environment");
     print "\nChanging $dummy $name to layout $layout" if $debug_on;
 
@@ -1331,6 +1406,11 @@ sub ConvertToLayout {
 	print OUTFILE "\n\\begin_deeper ";
     }
 
+    # If we still haven't printed the *previous* \layout command because that
+    # environment is empty, print it now! (This happens if an environment
+    # is nested inside a keepempty, like slide.)
+    &CheckForNewParagraph if $PendingLayout;
+
     # Put the new layout onto the layout stack
     push @$CurrentLayoutStack, $layout;
     print "\nCurrent Layout Stack: @$CurrentLayoutStack" if $debug_on;
@@ -1342,6 +1422,11 @@ sub ConvertToLayout {
     # Some environments can nest. Sections & Title commands can't
     $MayBeDeeper = $layoutref->{"nestable"};
 
+    # We haven't yet read any printable stuff in the new paragraph
+    # If this is a layout that's allowed to be empty, prepare for an
+    # empty paragraph
+    $PendingLayout = $keepempty;
+
 } # end sub ConvertToLayout
 
 sub EndLayout {
@@ -1358,6 +1443,12 @@ sub EndLayout {
     my $dummy = ($name =~ /^\\/ ? "macro" : "environment");
     print "\nEnding $dummy $name (layout $layout)" if $debug_on;
 
+    # If we still haven't printed the *previous* \layout command because that
+    # environment is empty, print it now! Before we pop the stack!
+    # This happens for a totally empty, non-nesting environment,
+    # like hollywood.sty's fadein
+    &CheckForNewParagraph if $PendingLayout;
+
     my $mylayout = pop (@$CurrentLayoutStack);
 
     # If a standard paragraph was the last thing in a list, then
